@@ -17,7 +17,9 @@
  * under the License.
  */
 
-#[derive(Clone, Debug, Default, Eq, PartialEq)]
+/// File stats that can be retrieved by reading the file's metadata.
+#[derive(Clone, Debug, Default)]
 pub struct FileStats {
     pub num_records: i64,
+    pub size_bytes: i64,
 }
